##simd level functions
floatn intel_sub_group_shuffle(floatn x, uint c)
intn intel_sub_group_shuffle(intn x, uint c)
uintn intel_sub_group_shuffle(uintn x, uint c)
floatn intel_sub_group_shuffle_down(floatn x, floatn y, uint c)
intn intel_sub_group_shuffle_down(intn x, intn y, uint c)
uintn intel_sub_group_shuffle_down(uintn x, uintn y, uint c)
floatn intel_sub_group_shuffle_up(floatn x, floatn y, uint c)
intn intel_sub_group_shuffle_up(intn x, intn y, uint c)
uintn intel_sub_group_shuffle_up(uintn x, uintn y, uint c)
floatn intel_sub_group_shuffle_xor(floatn x, uint c)
intn intel_sub_group_shuffle_xor(intn x, uint c)
uintn intel_sub_group_shuffle_xor(uintn x, uint c)
