##common
gentype clamp (gentype x, gentype minval, gentype maxval)
gentypef clamp (gentypef x, float minval, float maxval)
gentypeh clamp (gentypeh x, half minval, half maxval)
gentyped clamp (gentyped x, double minval, double maxval)
gentype degrees (gentype radians)
gentype max (gentype x,  gentype y)
gentypef max (gentypef x, float y)
gentypeh max (gentypeh x, half y)
gentyped max (gentyped x, double y)
gentype min (gentype x,  gentype y)
gentypef min (gentypef x,  float y)
gentypeh min (gentypeh x,  half y)
gentyped min (gentyped x,  double y)
gentype mix (gentype x, gentype y, gentype a)
gentypef mix (gentypef x, gentypef y, float a)
gentypeh mix (gentypeh x, gentypeh y, half a)
gentyped mix (gentyped x, gentyped y, double a)
gentype radians (gentype degrees)
gentype step (gentype edge, gentype x)
gentypef step (float edge, gentypef x)
gentypeh step (half edge, gentypeh x)
gentyped step (double edge, gentyped x)
gentype smoothstep (gentype edge0, gentype edge1, gentype x)
gentypef smoothstep (float edge0, float edge1, gentypef x)
gentypeh smoothstep (half edge0, half edge1, gentypeh x)
gentyped smoothstep (double edge0, double edge1, gentyped x)
gentype sign (gentype x)
