##relational
intn isequal (floatn x, floatn y)
shortn isequal (halfn x, halfn y)
longn isequal (doublen x, doublen y)
intn isnotequal (floatn x, floatn y)
shortn isnotequal (halfn x, halfn y)
longn isnotequal (doublen x, doublen y)
intn isgreater (floatn x, floatn y)
shortn isgreater (halfn x, halfn y)
longn isgreater (doublen x, doublen y)
intn isgreaterequal (floatn x, floatn y)
shortn isgreaterequal (halfn x, halfn y)
longn isgreaterequal (doublen x, doublen y)
intn isless (floatn x, floatn y)
shortn isless (halfn x, halfn y)
longn isless (doublen x, doublen y)
intn islessequal (floatn x, floatn y)
shortn islessequal (halfn x, halfn y)
longn islessequal (doublen x, doublen y)
intn islessgreater (floatn x, floatn y)
shortn islessgreater (halfn x, halfn y)
longn islessgreater (doublen x, doublen y)
intn isfinite (floatn)
shortn isfinite (halfn)
longn isfinite (doublen)
intn isinf (floatn)
shortn isinf (halfn)
longn isinf (doublen)
intn isnan (floatn)
shortn isnan (halfn)
longn isnan (doublen)
intn isnormal (floatn)
shortn isnormal (halfn)
longn isnormal (doublen)
intn isordered (floatn x, floatn y)
shortn isordered (halfn x, halfn y)
longn isordered (doublen x, doublen y)
intn isunordered (floatn x, floatn y)
shortn isunordered (halfn x, halfn y)
longn isunordered (doublen x, doublen y)
intn signbit (floatn)
shortn signbit (halfn)
longn signbit (doublen)
int any (igentype x)
int all (igentype x)
gentype bitselect (gentype a, gentype b, gentype c)
gentype select (gentype a, gentype b, igentype c)
gentype select (gentype a, gentype b, ugentype c)
